/** \file DisplayOpenGL.cpp
	\brief Display implementation based on OpenGL renderer
*/

/* Copyright (C) 2008-2009 Tomasz Ostrowski <tomasz.o.ostrowski at gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.

 * Miniscope is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <vcl.h>

#pragma hdrstop

#include "DisplayOpenGL.h"
#include "ValueStringFormatter.h"
#include <stdio.h>
#include <math.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <algorithm>


//---------------------------------------------------------------------------
#pragma package(smart_init)


void DisplayOpenGL::SetVsync(bool enabled)
{
    typedef BOOL (WINAPI * PFNWGLSWAPINTERVALEXTPROC) (int interval);
    PFNWGLSWAPINTERVALEXTPROC wglSwapInterval;
    wglSwapInterval = (PFNWGLSWAPINTERVALEXTPROC) wglGetProcAddress("wglSwapIntervalEXT");
    //sets the number of vert retraces before redrawing
    if ( wglSwapInterval )
        wglSwapInterval(enabled ? 1 : 0);
}

void DisplayOpenGL::CreateBmpFont(void)					 
{
	HFONT	font;						// Windows Font ID
	HFONT	oldfont;
	base = glGenLists(96);				// Storage For 96 Characters

	font = CreateFont(	-12,			// Height Of Font
		0,								// Width Of Font
		0,								// Angle Of Escapement
		0,								// Orientation Angle
		FW_BOLD,							// Font Weight
		FALSE,							// Italic
		FALSE,							// Underline
		FALSE,							// Strikeout
		ANSI_CHARSET,					// Character Set Identifier
		OUT_TT_PRECIS,					// Output Precision
		CLIP_DEFAULT_PRECIS,			// Clipping Precision
		ANTIALIASED_QUALITY,			// Output Quality
		FF_DONTCARE|DEFAULT_PITCH,		// Family And Pitch
		"Courier New");					// Font Name
	oldfont = (HFONT)SelectObject(this->hdc, font);
	wglUseFontBitmaps(this->hdc, 32, 96, base);
	SelectObject(this->hdc, oldfont);
	DeleteObject(font);
}

void DisplayOpenGL::DestroyFont(void)
{
	glDeleteLists(base, 96);	// delete char list
}

GLvoid DisplayOpenGL::Print(const char *fmt, ...)
{
	char buf[128];
	va_list	ap;
	if (fmt == NULL)
		return;

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf)-1, fmt, ap);
	va_end(ap);
	buf[sizeof(buf)-1] = NULL;

	glPushAttrib(GL_LIST_BIT);
	glListBase(base - 32);
	glCallLists(strlen(buf), GL_UNSIGNED_BYTE, buf);
    glPopAttrib();										
}


//---------------------------------------------------------------------------
DisplayOpenGL::DisplayOpenGL(void* hParent)
{
    iRange = 128;
	//_control87(MCW_EM, MCW_EM);
    SetVsync(true);
    hdc = GetDC(hParent);
    SetPixelFormatDescriptor();
    hrc = wglCreateContext(hdc);
    wglMakeCurrent(hdc, hrc);
	SetupRC();
	CreateBmpFont();

	scope.fZoomX = 1.0;
	scope.fZoomY = 1.0;
	scope.fCursor1X = 0.0;
	scope.fCursor1Y = 0.0;
	scope.fCursor2X = 0.0;
	scope.fCursor2Y = 0.0;
	scope.fOffsetX = 0.0;
	scope.fOffsetY = 0.0;
	spectrum.fZoomX = 1.0;
	spectrum.fZoomY = 1.0;
	spectrum.fCursor1X = 0.0;
	spectrum.fCursor1Y = 0.0;
	spectrum.fCursor2X = 0.0;
	spectrum.fCursor2Y = 0.0;
	spectrum.fOffsetX = 0.0;
	spectrum.fOffsetY = 0.0;

	data = NULL;
}
//---------------------------------------------------------------------------

DisplayOpenGL::~DisplayOpenGL()
{
    //kolejnosc?
    ReleaseDC(0, hdc);
    wglMakeCurrent(hdc, NULL);
    wglDeleteContext(hrc);
}


void DisplayOpenGL::ChangeSize(int iWidth, int iHeight)
{
//	fWidth = (float)iRange*iWidth/iHeight * 2.0f / State().fZoomX;
	fWidth = (float)iWidth / State().fZoomX;
	fHeight = (float)iRange / State().fZoomY;

	glViewport(0, 0, iWidth, iHeight);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
	glOrtho(State().fOffsetX,          	// left edge
		  State().fOffsetX+fWidth,     	// right edge
		  -fHeight+State().fOffsetY,   	// bottom edge
		  fHeight+State().fOffsetY,    	// top edge
		  0.0, 1.0);             		// near and far clipping, planes
    glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	this->iWidth = iWidth;
	this->iHeight = iHeight;
}

void DisplayOpenGL::SetMode(const enum E_DISPLAY_MODE m)
{
	mode = m;
	ChangeSize(iWidth, iHeight);
	Update(UPDATE_ALL);
}


#define MYGL_COLORS(color) GetRValue(color)/256.0, GetGValue(color)/256.0, GetBValue(color)/256.0
//---------------------------------------------------------------------------
void DisplayOpenGL::Update(enum E_UPDATE_LEVEL level)
{
	static TColor colorGrid = (TColor)0x303030;	// BGR
	static TColor colorLabel = (TColor)0x505050;
	static TColor colorCursor1 = (TColor)0x0000FF;
	static TColor colorCursor2 = (TColor)0xFF0000;

	float fVcoeff = fHeight * State().fZoomY / (float)iRange;		// float "canvas" per LSB

	glClear(GL_COLOR_BUFFER_BIT);

	glColor3f(MYGL_COLORS(colorGrid));
	glEnable(GL_LINE_STIPPLE);         /* dashed lines */
	glLineStipple(1, 0xF0F0);
	glBegin(GL_LINES);
	// horizontal grid
	for (int i=0; i<9; i++)
	{
		glVertex3f(State().fOffsetX, fHeight*(i-4)/5.0f + State().fOffsetY, 0.0f); // origin
		glVertex3f(fWidth+State().fOffsetX, fHeight*(i-4)/5.0f + State().fOffsetY, 0.0f); // end
	}
	// vertical grid
	for (int i=1; i<=9; i++)
	{
		glVertex3f(fWidth*(i)/10.0f + State().fOffsetX, -fHeight + State().fOffsetY, 0.0f); // origin
		glVertex3f(fWidth*(i)/10.0f + State().fOffsetX, fHeight + State().fOffsetY, 0.0f); // end
	}
	glEnd();
	glDisable(GL_LINE_STIPPLE);	

	glColor3f(MYGL_COLORS(colorLabel));
	// horizontal grid labels
	glColor3f(0.6f,0.6f,0.6f);
	std::string label;
	for (int i=-4; i<=4; i+=2)
	{
		glRasterPos2f(fWidth/40.0 + State().fOffsetX, fHeight*((float)i+0.07f)/5.0f + State().fOffsetY);
		float fVal = ((fHeight*((float)i)/5.0f) + State().fOffsetY) * fVcoeff * fDivY;
		label = ValueStringFormatter::VoltageToStr(fVal, 2);
		Print("%s", label.c_str());
	}

	// vertical grid labels
	for (int i=1; i<=9; i+=2)
	{
		glRasterPos2f(fWidth*((float)i+0.07f)/10.0f + State().fOffsetX, -fHeight*0.95f + State().fOffsetY);
		float fVal;
		if (mode == MODE_OSCILLOSCOPE)
		{
			fVal = (fWidth*((float)i)/10.0f + State().fOffsetX) * fDivX;
			label = ValueStringFormatter::TimeToStr(fVal, 2);
		}
		else if (mode == MODE_SPECTRUM)
		{
			if (data && data->size())
			{
				fVal = (fWidth*((float)i)/10.0f + State().fOffsetX) / (fDivX*data->size()*2.0f);
				label = ValueStringFormatter::FreqToStr(fVal, 2);
			}
		}
		Print("%s", label.c_str());
	}

	// cursors
	glColor3f(MYGL_COLORS(colorCursor1));
	glBegin(GL_LINES);
	glVertex3f(State().fCursor1X-fWidth/(2*10/* *(State().fZoomX/State().fZoomY)*/ *iWidth/iHeight), State().fCursor1Y, 0.0f);
	glVertex3f(State().fCursor1X+fWidth/(2*10/* *(State().fZoomX/State().fZoomY)*/ *iWidth/iHeight), State().fCursor1Y, 0.0f);
	glVertex3f(State().fCursor1X, State().fCursor1Y-fHeight/(10), 0.0f);
	glVertex3f(State().fCursor1X, State().fCursor1Y+fHeight/(10), 0.0f);
	glEnd();

	glColor3f(MYGL_COLORS(colorCursor2));
	glBegin(GL_LINES);
	glVertex3f(State().fCursor2X-fWidth/(2*10/* *(State().fZoomX/State().fZoomY)*/ *iWidth/iHeight), State().fCursor2Y, 0.0f);
	glVertex3f(State().fCursor2X+fWidth/(2*10/* *(State().fZoomX/State().fZoomY)*/ *iWidth/iHeight), State().fCursor2Y, 0.0f);
	glVertex3f(State().fCursor2X, State().fCursor2Y-fHeight/10, 0.0f);
	glVertex3f(State().fCursor2X, State().fCursor2Y+fHeight/10, 0.0f);
	glEnd();

	// plot
	if (data && data->size())
	{
		if (mode == MODE_OSCILLOSCOPE)
		{
			glColor3f(0.9f,0.9f,0.9f);
			glBegin(GL_LINE_STRIP);
			int i = std::max<float>(State().fOffsetX, 0.0f);
			float limit = std::min<float>(data->size(), fWidth + i);
			for (float fHPos = i; fHPos<limit; fHPos+=1.0f)
			{
				glVertex3f(fHPos, (*data)[i++]*fVcoeff, 0.0f);
			}
			glEnd();
		}
		else if (mode == MODE_SPECTRUM)
		{
			glColor3f(0.9f,0.9f,0.9f);
			int i = std::max<float>(State().fOffsetX, 0.0f);
			float limit = std::min<float>(data->size(), fWidth + i);
			glBegin(GL_LINES);
			for (float fHPos = i; fHPos<limit; fHPos+=1.0f)
			{
				glVertex3f(fHPos, 0.0f, 0.0f);
				glVertex3f(fHPos, (*data)[i++]*fVcoeff, 0.0f);
			}
			glEnd();

			i = std::max<float>(State().fOffsetX, 0.0f);
			glBegin(GL_POINT);
			for (float fHPos = i; fHPos<limit; fHPos+=1.0f)
			{
				if ((*data)[i] == 0.0f)
				{
					glVertex3f(fHPos, 0.0f, 0.0f);
				}
				i++;
			}
			glEnd();
		}
	}

	glFlush();
	SwapBuffers(hdc);
}
//---------------------------------------------------------------------------
void DisplayOpenGL::SetPixelFormatDescriptor()
{
	PIXELFORMATDESCRIPTOR pfd;
	ZeroMemory(&pfd, sizeof(pfd));					//ustalenie typu koloru, buforowania grafiki,
    pfd.nSize = sizeof(pfd);
    pfd.nVersion = 1;
    pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType = PFD_TYPE_RGBA;
    pfd.cColorBits = 24;
    pfd.cDepthBits = 16;
    pfd.iLayerType = PFD_MAIN_PLANE;
    int PixelFormat = ChoosePixelFormat(hdc, &pfd);
    SetPixelFormat(hdc, PixelFormat, &pfd);
}
//---------------------------------------------------------------------------
void DisplayOpenGL::SetupRC(void)
{
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);	// background color
	glClear(GL_COLOR_BUFFER_BIT);
	glFlush();
}

float DisplayOpenGL::FPosToPhysicalX(float fPosX)
{
	float fVal;
	if (mode == MODE_OSCILLOSCOPE)
		fVal = fPosX * fDivX;
	else
		fVal = fPosX / (fDivX*data->size()*2.0f);
	return fVal;
}

float DisplayOpenGL::FPosToPhysicalY(float fPosY)
{
	float fVal = fPosY * fDivY;
	return fVal;
}

float DisplayOpenGL::IPosToFloatX(int x)
{
	float val;
	val = State().fOffsetX + ((float)(x+1)/float(iWidth)*fWidth);
	return val;
}

float DisplayOpenGL::IPosToFloatY(int y)
{
	float val;
	val = State().fOffsetY +
		((float)(iHeight-y-1)/float(iHeight)*fHeight*2) - fHeight;
	return val;
}


void DisplayOpenGL::SetCursor1(int x, int y)
{
	State().fCursor1X = IPosToFloatX(x);
	State().fCursor1Y = IPosToFloatY(y);
}

void DisplayOpenGL::SetCursor2(int x, int y)
{
	State().fCursor2X = IPosToFloatX(x);
	State().fCursor2Y = IPosToFloatY(y);
}

void DisplayOpenGL::PosToPhysical(int x, int y, float &fValX, float &fValY)
{
	fValX = FPosToPhysicalX(IPosToFloatX(x));
	fValY = FPosToPhysicalY(IPosToFloatY(y));
}

void DisplayOpenGL::SetZoom(int iCenterX, int iCenterY, float fZoomX, float fZoomY)
{
	this->State().fZoomX = fZoomX;
	this->State().fZoomY = fZoomY;

	if (iCenterX == -1 && iCenterY == -1)
	{
		this->State().fOffsetX = 0;
		this->State().fOffsetY = 0;
	}
	else
	{
		this->State().fOffsetX = IPosToFloatX(iCenterX);
		this->State().fOffsetY = IPosToFloatY(iCenterY);
		//fWidth = (float)iRange*iWidth/iHeight * 2.0f / State().fZoomX;
		fWidth = (float)iWidth / State().fZoomX;
		//fHeight = (float)iRange / State().fZoomY;
		this->State().fOffsetX -= fWidth/2;
	}

	ChangeSize(this->iWidth, this->iHeight);
}

void DisplayOpenGL::Move(int iStartX, int iStartY, int iStopX, int iStopY)
{
	float deltaX = IPosToFloatX(iStartX) - IPosToFloatX(iStopX);
	float deltaY = IPosToFloatY(iStartY) - IPosToFloatY(iStopY);
	this->State().fOffsetX += deltaX;
	this->State().fOffsetY += deltaY;
	ChangeSize(this->iWidth, this->iHeight);
}

